/* Copyright (c) 2003 ossim.net
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *    from the author.
 *
 * 4. Products derived from this software may not be called "Os-sim" nor
 *    may "Os-sim" appear in their names without specific prior written
 *    permission from the author.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config.h>

#include <libgda/libgda.h>
#include <ossim.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

/* #include <popt.h> */

typedef struct {
  SimContainer   *container;
  SimConfig      *config;
} SimThreadData;

typedef struct {
  gchar          *config;
  gchar          *directive;
  gchar          *output;
} SimCmdArgs;

SimCmdArgs simCmdArgs;

/*
 *
 *
 *
 *
 */
static void
sim_log_handler (const gchar     *log_domain,
		 GLogLevelFlags   log_level,
		 const gchar     *message,
		 gpointer         data)
{
  gint fd;

  fd = open ("/tmp/ossim.log", O_CREAT|O_WRONLY|O_APPEND,  S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);

  write (fd, message, strlen(message));
  write (fd, "\n", 1);

  close (fd);
}

/*
 * sim_container_syslog:
 *
 *   arguments:
 *
 *   results:
 */

static gpointer
sim_thread_syslog (gpointer data)
{
  SimThreadData   *thr_data = (SimThreadData *) data;
  SimContainer    *container = thr_data->container;
  SimConfig       *config = thr_data->config;
  SimSyslog       *syslog;
  gchar           *log_file;
 
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (config);
  g_return_if_fail (SIM_IS_CONFIG (config));

  g_message ("sim_thread_syslog");

  log_file = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_LOG_FILE);

  syslog = sim_syslog_new (container, log_file);
  g_assert (syslog != NULL);
  sim_syslog_run (syslog);

  return NULL;
}

/*
 * sim_container_scheduler:
 *
 *   arguments:
 *
 *   results:
 */
static gpointer
sim_thread_scheduler (gpointer data)
{
  SimThreadData   *thr_data = (SimThreadData *) data;
  SimContainer    *container = thr_data->container;
  SimConfig       *config = thr_data->config;
  SimDatabase     *database;
  SimScheduler    *scheduler;
  gchar           *datasource;
  gchar           *username;
  gchar           *password;
 
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (config);
  g_return_if_fail (SIM_IS_CONFIG (config));

  g_message ("sim_thread_scheduler");

  datasource = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_DATABASE);
  username = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_USERNAME);
  password = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_PASSWORD);
  database = sim_database_new (datasource, username, password);

  scheduler = sim_scheduler_new (container, database, config);
  sim_scheduler_run (scheduler);

  return NULL;
}

/**
 * sim_container_organizer:
 *
 *   arguments:
 *
 *   results:
 */

static gpointer
sim_thread_organizer (gpointer data)
{
  SimThreadData   *thr_data = (SimThreadData *) data;
  SimContainer    *container = thr_data->container;
  SimConfig       *config = thr_data->config;
  SimDatabase     *database;
  SimOrganizer    *organizer;
  gchar           *datasource;
  gchar           *username;
  gchar           *password;
 
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (config);
  g_return_if_fail (SIM_IS_CONFIG (config));

  g_message ("sim_thread_organizer");

  datasource = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_DATABASE);
  username = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_USERNAME);
  password = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_PASSWORD);
  database = sim_database_new (datasource, username, password);

  organizer = sim_organizer_new (container, database);
  sim_organizer_run (organizer);

  return NULL;
}

/**
 * sim_container_organizer:
 *
 *   arguments:
 *
 *   results:
 */

static gpointer
sim_thread_server (gpointer data)
{
  SimThreadData   *thr_data = (SimThreadData *) data;
  SimContainer    *container = thr_data->container;
  SimConfig       *config = thr_data->config;
  SimDatabase     *database;
  SimServer       *server;
  gchar           *datasource;
  gchar           *username;
  gchar           *password;
 
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (config);
  g_return_if_fail (SIM_IS_CONFIG (config));

  g_message ("sim_thread_server");

  datasource = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_DATABASE);
  username = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_USERNAME);
  password = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_PASSWORD);
  database = sim_database_new (datasource, username, password);

  server = sim_server_new (container, database);
  sim_server_run (server);

  return NULL;
}

/*
 *
 *
 *
 */
static void
options (int argc, const char **argv)
{
  /*
  poptContext     context;


  static const struct poptOption options[] ={
    {
      "config", 
      'c', 
      POPT_ARG_STRING, 
      &simCmdArgs.config,
      0,
      "source file configuration default: config.xml", 
      "filename"
    },
    {
      "directives", 
      'd', 
      POPT_ARG_STRING, 
      &simCmdArgs.directive,
      0,
      "source file directives default: directives.xml", 
      "filename"
    },
    { 
      "output", 
      'o', 
      POPT_ARG_STRING, 
      &simCmdArgs.output,
      0,
      "send query results to file default: ossim.log", 
      "filename"
    },
    POPT_AUTOHELP
    { NULL, '\0', 0, NULL, 0 }
  };

  context = poptGetContext (NULL, argc, argv, options, 0);
  if (!context)
    {
      g_print ("Error: Invalid command line.\n");
      return;
    }

  if ((poptGetNextOpt (context) < -1) || (poptGetArg (context) != NULL))
    {
      g_print ("Error:Incorrect argument in command line.\n");
      poptPrintHelp (context, stderr, 0);
      poptFreeContext(context);
      exit (1);
    }

  poptFreeContext(context);
  */
}


/*
 *
 *
 *
 */
int
main (int argc, char *argv[])
{
  GMainLoop     *loop;
  SimThreadData *data;
  SimConfig     *config; 
  SimContainer  *container;
  SimDatabase   *database;
  GThread       *thread;
  gchar         *datasource;
  gchar         *username;
  gchar         *password;

  /* Default Command Line Options */
  simCmdArgs.config = NULL;
  simCmdArgs.directive = NULL;
  simCmdArgs.output = NULL;

  /* Command Line Options */
  options (argc, (const char **) argv);

  /* Thread Init */
  if (!g_thread_supported ()) 
    {
      g_thread_init (NULL);
    }

  gda_init (NULL, "0.1", argc, argv);

  g_log_set_handler (NULL,
		     G_LOG_LEVEL_WARNING | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_MESSAGE, 
		     sim_log_handler, 
		     NULL);
  
  g_log_set_handler (G_LOG_DOMAIN, 
		     G_LOG_LEVEL_WARNING | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_MESSAGE, 
		     sim_log_handler, 
		     NULL);
  
  
  /* Config Init */
  config = sim_config_new (SIM_CONFIG_FILE);

  /* Database init */
  datasource = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_DATABASE);
  username = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_USERNAME);
  password = sim_config_get_property_value (config, SIM_CONFIG_PROPERTY_TYPE_PASSWORD);
  database = sim_database_new (datasource, username, password);

  /* Container init */
  container = sim_container_new ();
  sim_container_db_load_hosts (container, database);
  sim_container_db_load_nets (container, database);
  sim_container_db_load_signatures (container, database);
  sim_container_db_load_policies (container, database);
  sim_container_db_load_host_levels (container, database);
  sim_container_db_load_net_levels (container, database);
  sim_container_load_directives_from_file (container, SIM_XML_DIRECTIVE_FILE);

  g_object_unref (database);

  /* Thread Data Init */
  data = g_new0 (SimThreadData, 1);
  data->container = container;
  data->config = config;

  /* Syslog Thread */
  thread = g_thread_create (sim_thread_syslog, data, TRUE, NULL);
  g_return_if_fail (thread);
  g_thread_set_priority (thread, G_THREAD_PRIORITY_NORMAL);

  /* Scheduler Thread */
  thread = g_thread_create (sim_thread_scheduler, data, TRUE, NULL);
  g_return_if_fail (thread);
  g_thread_set_priority (thread, G_THREAD_PRIORITY_NORMAL);

  /* Organizer Thread */
  thread = g_thread_create (sim_thread_organizer, data, TRUE, NULL);
  g_return_if_fail (thread);
  g_thread_set_priority (thread, G_THREAD_PRIORITY_NORMAL);

  /* Server Thread */
  thread = g_thread_create (sim_thread_server, data, TRUE, NULL);
  g_return_if_fail (thread);
  g_thread_set_priority (thread, G_THREAD_PRIORITY_NORMAL);

  /* Main Loop */
  loop = g_main_loop_new (NULL, FALSE);
  g_main_loop_run (loop);

  return 0;
}
