/* Copyright (c) 2003 ossim.net
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *    from the author.
 *
 * 4. Products derived from this software may not be called "Os-sim" nor
 *    may "Os-sim" appear in their names without specific prior written
 *    permission from the author.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <config.h>

#include "sim-container.h"
#include "sim-xml-directive.h"

G_LOCK_DEFINE_STATIC (s_mutex_config);
G_LOCK_DEFINE_STATIC (s_mutex_hosts);
G_LOCK_DEFINE_STATIC (s_mutex_nets);
G_LOCK_DEFINE_STATIC (s_mutex_signatures);
G_LOCK_DEFINE_STATIC (s_mutex_policies);
G_LOCK_DEFINE_STATIC (s_mutex_directives);
G_LOCK_DEFINE_STATIC (s_mutex_host_levels);
G_LOCK_DEFINE_STATIC (s_mutex_net_levels);
G_LOCK_DEFINE_STATIC (s_mutex_backlogs);
G_LOCK_DEFINE_STATIC (s_mutex_messages);

enum
{
  DESTROY,
  LAST_SIGNAL
};

struct _SimContainerPrivate {
  GList        *hosts;
  GList        *nets;
  GList        *signatures;
  GList        *policies;
  GList        *directives;

  GList        *host_levels;
  GList        *net_levels;
  GList        *backlogs;

  GQueue       *messages;

  GCond        *cond_messages;
  GMutex       *mutex_messages;
};

static gpointer parent_class = NULL;
static gint sim_container_signals[LAST_SIGNAL] = { 0 };

/* GType Functions */

static void 
sim_container_impl_dispose (GObject  *gobject)
{
  G_OBJECT_CLASS (parent_class)->dispose (gobject);
}

static void
sim_container_impl_finalize (GObject  *gobject)
{
  SimContainer  *container = SIM_CONTAINER (gobject);

  sim_container_free_hosts_ul (container);
  sim_container_free_nets_ul (container);
  sim_container_free_signatures_ul (container);
  sim_container_free_policies_ul (container);
  sim_container_free_directives_ul (container);
  sim_container_free_host_levels_ul (container);
  sim_container_free_net_levels_ul (container);
  sim_container_free_backlogs_ul (container);
  sim_container_free_messages (container);

  g_cond_free (container->_priv->cond_messages);
  g_mutex_free (container->_priv->mutex_messages);

  g_free (container->_priv);

  G_OBJECT_CLASS (parent_class)->finalize (gobject);
}

static void
sim_container_class_init (SimContainerClass * class)
{
  GObjectClass *object_class = G_OBJECT_CLASS (class);

  parent_class = g_type_class_ref (G_TYPE_OBJECT);

  object_class->dispose = sim_container_impl_dispose;
  object_class->finalize = sim_container_impl_finalize;
}

static void
sim_container_instance_init (SimContainer *container)
{
  container->_priv = g_new0 (SimContainerPrivate, 1);

  container->_priv->hosts = NULL;
  container->_priv->nets = NULL;
  container->_priv->signatures = NULL;

  container->_priv->policies = NULL;
  container->_priv->directives = NULL;

  container->_priv->host_levels = NULL;
  container->_priv->net_levels = NULL;
  container->_priv->backlogs = NULL;

  container->_priv->messages = g_queue_new ();

  /* Mutex Messages Init */
  container->_priv->cond_messages = g_cond_new ();
  container->_priv->mutex_messages = g_mutex_new ();
}

/* Public Methods */

GType
sim_container_get_type (void)
{
  static GType object_type = 0;
 
  if (!object_type)
  {
    static const GTypeInfo type_info = {
              sizeof (SimContainerClass),
              NULL,
              NULL,
              (GClassInitFunc) sim_container_class_init,
              NULL,
              NULL,                       /* class data */
              sizeof (SimContainer),
              0,                          /* number of pre-allocs */
              (GInstanceInitFunc) sim_container_instance_init,
              NULL                        /* value table */
    };
    
    g_type_init ();
                                                                                                                             
    object_type = g_type_register_static (G_TYPE_OBJECT, "SimContainer", &type_info, 0);
  }

  return object_type;
}

/*
 *
 *
 *
 *
 */
SimContainer*
sim_container_new (void)
{
  SimContainer *container = NULL;

  container = SIM_CONTAINER (g_object_new (SIM_TYPE_CONTAINER, NULL));

  return container;
}

/*
 *
 *
 *
 *
 */
gint
sim_container_db_get_recovery_ul (SimContainer  *container,
				  SimDatabase   *database)
{
  GdaDataModel  *dm;
  GdaValue      *value;
  gchar         *query = "SELECT recovery FROM conf";
  gint           row;
  gint           recovery = 1;

  g_return_if_fail (container != NULL);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database != NULL);
  g_return_if_fail (SIM_IS_DATABASE (database));
  
  dm = sim_database_execute_single_command (database, query);
  if (dm)
    {
      for (row = 0; row < gda_data_model_get_n_rows (dm); row++)
	{
	  /* Recovery */
	  value = (GdaValue *) gda_data_model_get_value_at (dm, 0, row);
	  recovery = gda_value_get_integer (value);
	}
      
      g_object_unref(dm);
    }
  else
    {
      g_message ("RECOVERY DATA MODEL ERROR");
    }

  return recovery;
}

/*
 *
 *
 *
 *
 */
gint
sim_container_db_get_recovery (SimContainer  *container,
			       SimDatabase   *database)
{
  gint   recovery;

  g_return_if_fail (container != NULL);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database != NULL);
  g_return_if_fail (SIM_IS_DATABASE (database));

  G_LOCK (s_mutex_config);
  recovery = sim_container_db_get_recovery_ul (container, database);
  G_UNLOCK (s_mutex_config);

  return recovery;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_hosts_ul (SimContainer  *container,
				SimDatabase   *database)
{
  SimHost       *host;
  GdaDataModel  *dm;
  gint           row;
  gchar         *query = "SELECT * FROM host";

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  dm = sim_database_execute_single_command (database, query);
  if (dm)
    {
      for (row = 0; row < gda_data_model_get_n_rows (dm); row++)
	{
	  host  = sim_host_new_from_dm (dm, row);
	  container->_priv->hosts = g_list_append (container->_priv->hosts, host);
	}

      g_object_unref(dm);
    }
  else
    {
      g_message ("HOSTS DATA MODEL ERROR");
    }
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_host_ul (SimContainer  *container,
			      SimHost       *host)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host);
  g_return_if_fail (SIM_IS_HOST (host));

  container->_priv->hosts = g_list_append (container->_priv->hosts, host);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_host_ul (SimContainer  *container,
			      SimHost       *host)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host);
  g_return_if_fail (SIM_IS_HOST (host));

  container->_priv->hosts = g_list_remove (container->_priv->hosts, host);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_hosts_ul (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->hosts);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_hosts_ul (SimContainer  *container,
			    GList         *hosts)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (hosts);

  container->_priv->hosts = g_list_concat (container->_priv->hosts, hosts);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_hosts_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->hosts;
  while (list)
    {
      SimHost *host = (SimHost *) list->data;
      g_object_unref (host);

      list = list->next;
    }
  g_list_free (container->_priv->hosts);
}

/*
 *
 *
 *
 *
 */
SimHost*
sim_container_get_host_by_ia_ul (SimContainer  *container,
				 GInetAddr     *ia)
{
  SimHost   *host;
  GList     *list;
  gboolean   found = FALSE;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (ia, NULL);

  list = container->_priv->hosts;
  while (list)
    {
      host = (SimHost *) list->data;

      if (gnet_inetaddr_noport_equal (sim_host_get_ia (host), ia))
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return NULL;

  return host;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_hosts (SimContainer  *container,
			     SimDatabase   *database)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  G_LOCK (s_mutex_hosts);
  sim_container_db_load_hosts_ul (container, database);
  G_UNLOCK (s_mutex_hosts);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_host (SimContainer  *container,
			   SimHost       *host)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host);
  g_return_if_fail (SIM_IS_HOST (host));

  G_LOCK (s_mutex_hosts);
  sim_container_append_host_ul (container, host);
  G_UNLOCK (s_mutex_hosts);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_host (SimContainer  *container,
			   SimHost       *host)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host);
  g_return_if_fail (SIM_IS_HOST (host));

  G_LOCK (s_mutex_hosts);
  sim_container_remove_host_ul (container, host);
  G_UNLOCK (s_mutex_hosts);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_hosts (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_hosts);
  list = sim_container_get_hosts_ul (container);
  G_UNLOCK (s_mutex_hosts);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_hosts (SimContainer  *container,
			 GList         *hosts)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (hosts);

  G_LOCK (s_mutex_hosts);
  sim_container_set_hosts_ul (container, hosts);
  G_UNLOCK (s_mutex_hosts);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_hosts (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_hosts);
  sim_container_free_hosts_ul (container);
  G_UNLOCK (s_mutex_hosts);
}

/*
 *
 *
 *
 *
 */
SimHost*
sim_container_get_host_by_ia (SimContainer  *container,
			      GInetAddr     *ia)
{
  SimHost   *host;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (ia, NULL);

  G_LOCK (s_mutex_hosts);
  host = sim_container_get_host_by_ia_ul (container, ia);
  G_UNLOCK (s_mutex_hosts);

  return host;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_nets_ul (SimContainer  *container,
			       SimDatabase   *database)
{
  SimNet        *net;
  GdaDataModel  *dm;
  GdaDataModel  *dm2;
  GdaValue      *value;
  GInetAddr     *ia;
  gint           row;
  gint           row2;
  gchar         *query = "SELECT * FROM net";
  gchar         *query2;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  dm = sim_database_execute_single_command (database, query);
  if (dm)
    {
      for (row = 0; row < gda_data_model_get_n_rows (dm); row++)
	{
	  net  = sim_net_new_from_dm (dm, row);

	  query2 = g_strdup_printf ("SELECT host_ip FROM net_host_reference WHERE net_name = '%s'",
				    sim_net_get_name (net));

	  dm2 = sim_database_execute_single_command (database, query2);
	  if (dm2)
	    {
	      for (row2 = 0; row2 < gda_data_model_get_n_rows (dm2); row2++)
		{
		  gchar *ip;

		  value = (GdaValue *) gda_data_model_get_value_at (dm2, 0, row2);
		  ip = gda_value_stringify (value);

		  ia = gnet_inetaddr_new_nonblock (ip, 0);
		  sim_net_append_ia (net, ia);

		  g_free (ip);
		}
	      g_object_unref(dm2);
	    }
	  else
	    {
	      g_message ("NET HOST REFERENCES DATA MODEL ERROR");
	    }

	  g_free (query2);

	  container->_priv->nets = g_list_append (container->_priv->nets, net);
	}
      
      g_object_unref(dm);
    }
  else
    {
      g_message ("NETS DATA MODEL ERROR");
    }
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_net_ul (SimContainer  *container,
			     SimNet        *net)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net);
  g_return_if_fail (SIM_IS_NET (net));

  container->_priv->nets = g_list_append (container->_priv->nets, net);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_net_ul (SimContainer  *container,
			     SimNet        *net)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net);
  g_return_if_fail (SIM_IS_NET (net));

  container->_priv->nets = g_list_remove (container->_priv->nets, net);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_nets_ul (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->nets);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_nets_ul (SimContainer  *container,
			   GList         *nets)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (nets);

  container->_priv->nets = g_list_concat (container->_priv->nets, nets);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_nets_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->nets;
  while (list)
    {
      SimNet *net = (SimNet *) list->data;
      g_object_unref (net);

      list = list->next;
    }
  g_list_free (container->_priv->nets);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_nets_has_ia_ul (SimContainer  *container,
				  GInetAddr     *ia)
{
  GList *list;
  GList *nets = NULL;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (ia, NULL);

  list = container->_priv->nets;
  while (list)
    {
      SimNet *net = (SimNet *) list->data;

      if (sim_net_has_ia (net, ia))
	{
	  nets = g_list_append (nets, net);
	}

      list = list->next;
    }

  return nets;
}

/*
 *
 *
 *
 *
 */
SimNet*
sim_container_get_net_by_name_ul (SimContainer  *container,
				  const gchar   *name)
{
  SimNet    *net;
  GList     *list;
  gboolean   found = FALSE;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (name, NULL);

  list = container->_priv->nets;
  while (list)
    {
      net = (SimNet *) list->data;

      if (!strcmp (sim_net_get_name (net), name))
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return NULL;

  return net;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_nets (SimContainer  *container,
			    SimDatabase   *database)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  G_LOCK (s_mutex_nets);
  sim_container_db_load_nets_ul (container, database);
  G_UNLOCK (s_mutex_nets);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_net (SimContainer  *container,
			  SimNet        *net)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net);
  g_return_if_fail (SIM_IS_NET (net));

  G_LOCK (s_mutex_nets);
  sim_container_append_net_ul (container, net);
  G_UNLOCK (s_mutex_nets);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_net (SimContainer  *container,
			  SimNet        *net)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net);
  g_return_if_fail (SIM_IS_NET (net));

  G_LOCK (s_mutex_nets);
  sim_container_remove_net_ul (container, net);
  G_UNLOCK (s_mutex_nets);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_nets (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_nets);
  list = sim_container_get_nets_ul (container);
  G_UNLOCK (s_mutex_nets);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_nets (SimContainer  *container,
			GList         *nets)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (nets);

  G_LOCK (s_mutex_nets);
  sim_container_set_nets_ul (container, nets);
  G_UNLOCK (s_mutex_nets);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_nets (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_nets);
  sim_container_free_nets_ul (container);
  G_UNLOCK (s_mutex_nets);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_nets_has_ia (SimContainer  *container,
			       GInetAddr     *ia)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (ia, NULL);

  G_LOCK (s_mutex_nets);
  list = sim_container_get_nets_has_ia_ul (container, ia);
  G_UNLOCK (s_mutex_nets);

  return list;
}

/*
 *
 *
 *
 *
 */
SimNet*
sim_container_get_net_by_name (SimContainer  *container,
			       const gchar   *name)
{
  SimNet    *net;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (name, NULL);

  G_LOCK (s_mutex_nets);
  net = sim_container_get_net_by_name_ul (container, name);
  G_UNLOCK (s_mutex_nets);

  return net;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_signatures_ul (SimContainer  *container,
				     SimDatabase   *database)
{
  SimSignature   *signature;
  SimSignature   *child;
  GdaDataModel   *dm;
  GdaValue       *value;
  gchar          *query = "select * from signature";
  gint            row;
  FILE           *fd;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  dm = sim_database_execute_single_command (database, query);
  if (dm)
    {
      for (row = 0; row < gda_data_model_get_n_rows (dm); row++)
	{
	  signature = sim_signature_new_from_dm (SIM_SIGNATURE_TYPE_GROUP, dm, row);

	  container->_priv->signatures = g_list_append (container->_priv->signatures, signature);
	}
    }
  else
    {
      g_message ("SIGNATURE DATA MODEL ERROR");
    }

  if ((fd = fopen ("sids", "r")) == NULL)
    {
      printf ("Can't open file %s\n", "sids");
      exit (-1);
    }
  
  while (!feof (fd))
    {
      gchar str[BUFFER_SIZE], *ret, **split;
      gint id;

      ret = fgets (str, BUFFER_SIZE, fd);
      
      split = g_strsplit (str, ";", 0);
      
      id = strtol(split[1], (char **)NULL, 10);

      signature = sim_container_get_signature_group_by_name_ul (container, split[0]);
      child = sim_signature_new (SIM_SIGNATURE_TYPE_SIGNATURE, id, split[2]);      
      sim_signature_append_child (signature, child);
      
      g_strfreev (split);
    }

  fclose (fd);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_signature_ul (SimContainer  *container,
				   SimSignature  *signature)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (signature);
  g_return_if_fail (SIM_IS_SIGNATURE (signature));

  container->_priv->signatures = g_list_append (container->_priv->signatures, signature);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_signature_ul (SimContainer  *container,
				   SimSignature  *signature)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (signature);
  g_return_if_fail (SIM_IS_SIGNATURE (signature));

  container->_priv->signatures = g_list_remove (container->_priv->signatures, signature);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_signatures_ul (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->signatures);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_signatures_ul (SimContainer  *container,
				 GList         *signatures)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (signatures);

  container->_priv->signatures = g_list_concat (container->_priv->signatures, signatures);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_signatures_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->signatures;
  while (list)
    {
      SimSignature *signature = (SimSignature *) list->data;
      g_object_unref (signature);

      list = list->next;
    }
  g_list_free (container->_priv->signatures);
}

/*
 *
 *
 *
 *
 */
SimSignature*
sim_container_get_signature_group_by_id_ul (SimContainer  *container,
					    gint           id)
{
  SimSignature  *signature;
  GList         *list;
  gboolean       found = FALSE;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  
  list = container->_priv->signatures;
  while (list)
    {
      signature = (SimSignature *) list->data;

      if (sim_signature_get_id (signature) == id)
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return NULL;

  return signature;
}

/*
 *
 *
 *
 *
 */
SimSignature*
sim_container_get_signature_group_by_name_ul (SimContainer  *container,
					      const gchar   *name)
{
  SimSignature  *signature;
  GList         *list;
  gboolean       found = FALSE;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (name, NULL);
  
  list = container->_priv->signatures;
  while (list)
    {
      signature = (SimSignature *) list->data;

      if (!strcmp (sim_signature_get_name (signature), name))
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return NULL;

  return signature;
}

/*
 *
 *
 *
 *
 */
SimSignature*
sim_container_get_signature_group_by_sid_ul (SimContainer  *container,
					     gint           sid)
{
  SimSignature  *signature;
  GList         *list;
  gboolean       found = FALSE;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = container->_priv->signatures;
  while (list)
    {
      signature = (SimSignature *) list->data;

      if (sim_signature_has_child_id (signature, sid))
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return NULL;

  return signature;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_signatures (SimContainer  *container,
				  SimDatabase   *database)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  G_LOCK (s_mutex_signatures);
  sim_container_db_load_signatures_ul (container, database);
  G_UNLOCK (s_mutex_signatures);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_signature (SimContainer  *container,
				SimSignature  *signature)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (signature);
  g_return_if_fail (SIM_IS_SIGNATURE (signature));

  G_LOCK (s_mutex_signatures);
  sim_container_append_signature_ul (container, signature);
  G_UNLOCK (s_mutex_signatures);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_signature (SimContainer  *container,
				SimSignature  *signature)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (signature);
  g_return_if_fail (SIM_IS_SIGNATURE (signature));

  G_LOCK (s_mutex_signatures);
  sim_container_remove_signature_ul (container, signature);
  G_UNLOCK (s_mutex_signatures);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_signatures (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_signatures);
  list = sim_container_get_signatures_ul (container);
  G_UNLOCK (s_mutex_signatures);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_signatures (SimContainer  *container,
			      GList         *signatures)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (signatures);

  G_LOCK (s_mutex_signatures);
  sim_container_set_signatures_ul (container, signatures);
  G_UNLOCK (s_mutex_signatures);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_signatures (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_signatures);
  sim_container_free_signatures_ul (container);
  G_UNLOCK (s_mutex_signatures);
}

/*
 *
 *
 *
 *
 */
SimSignature*
sim_container_get_signature_group_by_id (SimContainer  *container,
					  gint          id)
{
  SimSignature  *signature;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  
  G_LOCK (s_mutex_signatures);
  signature = sim_container_get_signature_group_by_id_ul (container, id);
  G_UNLOCK (s_mutex_signatures);

  return signature;
}

/*
 *
 *
 *
 *
 */
SimSignature*
sim_container_get_signature_group_by_name (SimContainer  *container,
					   const gchar   *name)
{
  SimSignature  *signature;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (name, NULL);
  
  G_LOCK (s_mutex_signatures);
  signature = sim_container_get_signature_group_by_name_ul (container, name);
  G_UNLOCK (s_mutex_signatures);

  return signature;
}

/*
 *
 *
 *
 *
 */
SimSignature*
sim_container_get_signature_group_by_sid (SimContainer  *container,
					  gint           sid)
{
  SimSignature  *signature;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_signatures);
  signature = sim_container_get_signature_group_by_sid_ul (container, sid);
  G_UNLOCK (s_mutex_signatures);

  return signature;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_policies_ul (SimContainer  *container,
				   SimDatabase   *database)
{
  SimPolicy     *policy;
  GdaDataModel  *dm;
  GdaDataModel  *dm2;
  GdaValue      *value;
  GInetAddr     *ia;
  gint           row;
  gint           row2;
  gchar         *query = "SELECT policy.id, policy.priority, policy.descr, policy_time.begin_hour, policy_time.end_hour, policy_time.begin_day, policy_time.end_day FROM policy, policy_time WHERE policy.id = policy_time.policy_id;";
  gchar         *query2;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  dm = sim_database_execute_single_command (database, query);
  if (dm)
    {
      for (row = 0; row < gda_data_model_get_n_rows (dm); row++)
	{
	  policy  = sim_policy_new_from_dm (dm, row);

	  /* Host Source Inet Address */
	  query2 = g_strdup_printf ("SELECT host_ip FROM  policy_host_reference WHERE policy_id = %d AND direction = 'source'",
				    sim_policy_get_id (policy));
	  dm2 = sim_database_execute_single_command (database, query2);
	  if (dm2)
	    {
	      for (row2 = 0; row2 < gda_data_model_get_n_rows (dm2); row2++)
		{
		  gchar *src_ip;

		  value = (GdaValue *) gda_data_model_get_value_at (dm2, 0, row2);
		  src_ip = gda_value_stringify (value);

		  if (!g_ascii_strncasecmp (src_ip, SIM_IN_ADDR_ANY_CONST, 3))
		    ia = gnet_inetaddr_new_nonblock (SIM_IN_ADDR_ANY_IP_STR, 0);
		  else
		    ia = gnet_inetaddr_new_nonblock (src_ip, 0);

		  sim_policy_append_src_ia (policy, ia);

		  g_free (src_ip);
		}
	      g_object_unref(dm2);
	    }
	  else
	    {
	      g_message ("POLICY HOST SOURCE REFERENCES DATA MODEL ERROR");
	    }
	  g_free (query2);

	  /* Host Destination Inet Address */
	  query2 = g_strdup_printf ("SELECT host_ip FROM  policy_host_reference WHERE policy_id = %d AND direction = 'dest'",
				    sim_policy_get_id (policy));
	  dm2 = sim_database_execute_single_command (database, query2);
	  if (dm2)
	    {
	      for (row2 = 0; row2 < gda_data_model_get_n_rows (dm2); row2++)
		{
		  gchar *dst_ip;

		  value = (GdaValue *) gda_data_model_get_value_at (dm2, 0, row2);
		  dst_ip = gda_value_stringify (value);

		  if (!g_ascii_strncasecmp (dst_ip, SIM_IN_ADDR_ANY_CONST, 3))
		    ia = gnet_inetaddr_new_nonblock (SIM_IN_ADDR_ANY_IP_STR, 0);
		  else
		    ia = gnet_inetaddr_new_nonblock (dst_ip, 0);

		  sim_policy_append_dst_ia (policy, ia);

		  g_free (dst_ip);
		}
	      g_object_unref(dm2);
	    }
	  else
	    {
	      g_message ("POLICY HOST DEST REFERENCES DATA MODEL ERROR");
	    }
	  g_free (query2);


	  /* Net Source Inet Address */
	  query2 = g_strdup_printf ("SELECT host_ip FROM policy_net_reference, net_host_reference WHERE policy_net_reference.net_name = net_host_reference.net_name AND policy_net_reference.direction = 'source' AND policy_id = %d",
				    sim_policy_get_id (policy));
	  dm2 = sim_database_execute_single_command (database, query2);
	  if (dm2)
	    {
	      for (row2 = 0; row2 < gda_data_model_get_n_rows (dm2); row2++)
		{
		  gchar *src_ip;

		  value = (GdaValue *) gda_data_model_get_value_at (dm2, 0, row2);
		  src_ip = gda_value_stringify (value);

		  if (!g_ascii_strncasecmp (src_ip, SIM_IN_ADDR_ANY_CONST, 3))
		    ia = gnet_inetaddr_new_nonblock (SIM_IN_ADDR_ANY_IP_STR, 0);
		  else
		    ia = gnet_inetaddr_new_nonblock (src_ip, 0);

		  sim_policy_append_src_ia (policy, ia);

		  g_free (src_ip);
		}
	      g_object_unref(dm2);
	    }
	  else
	    {
	      g_message ("POLICY NET SOURCE REFERENCES DATA MODEL ERROR");
	    }
	  g_free (query2);


	  /* Net Destination Inet Address */
	  query2 = g_strdup_printf ("SELECT host_ip FROM policy_net_reference, net_host_reference WHERE policy_net_reference.net_name = net_host_reference.net_name AND policy_net_reference.direction = 'dest' AND policy_id = %d",
				    sim_policy_get_id (policy));
	  dm2 = sim_database_execute_single_command (database, query2);
	  if (dm2)
	    {
	      for (row2 = 0; row2 < gda_data_model_get_n_rows (dm2); row2++)
		{
		  gchar *dst_ip;

		  value = (GdaValue *) gda_data_model_get_value_at (dm2, 0, row2);
		  dst_ip = gda_value_stringify (value);

		  if (!g_ascii_strncasecmp (dst_ip, SIM_IN_ADDR_ANY_CONST, 3))
		    ia = gnet_inetaddr_new_nonblock (SIM_IN_ADDR_ANY_IP_STR, 0);
		  else
		    ia = gnet_inetaddr_new_nonblock (dst_ip, 0);

		  sim_policy_append_dst_ia (policy, ia);

		  g_free (dst_ip);
		}
	      g_object_unref(dm2);
	    }
	  else
	    {
	      g_message ("POLICY NET DEST REFERENCES DATA MODEL ERROR");
	    }
	  g_free (query2);

	  /* Signatures */
	  query2 = g_strdup_printf ("SELECT sig_name FROM policy_sig_reference, signature_group_reference WHERE policy_sig_reference.sig_group_name = signature_group_reference.sig_group_name AND policy_sig_reference.policy_id = %d",
				    sim_policy_get_id (policy));
	  dm2 = sim_database_execute_single_command (database, query2);
	  if (dm2)
	    {
	      for (row2 = 0; row2 < gda_data_model_get_n_rows (dm2); row2++)
		{
		  gchar *signature;

		  value = (GdaValue *) gda_data_model_get_value_at (dm2, 0, row2);
		  signature = gda_value_stringify (value);

		  sim_policy_append_signature (policy, signature);

		  g_free (signature);
		}
	      g_object_unref(dm2);
	    }
	  else
	    {
	      g_message ("POLICY SIGNATURE REFERENCES DATA MODEL ERROR");
	    }
	  g_free (query2);

	  container->_priv->policies = g_list_append (container->_priv->policies, policy);
	}
      g_object_unref(dm);
    }
  else
    {
      g_message ("POLICY DATA MODEL ERROR");
    }
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_policy_ul (SimContainer  *container,
				SimPolicy       *policy)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (policy);
  g_return_if_fail (SIM_IS_POLICY (policy));

  container->_priv->policies = g_list_append (container->_priv->policies, policy);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_policy_ul (SimContainer  *container,
				SimPolicy       *policy)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (policy);
  g_return_if_fail (SIM_IS_POLICY (policy));

  container->_priv->policies = g_list_remove (container->_priv->policies, policy);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_policies_ul (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->policies);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_policies_ul (SimContainer  *container,
			       GList         *policies)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (policies);

  container->_priv->policies = g_list_concat (container->_priv->policies, policies);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_policies_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->policies;
  while (list)
    {
      SimPolicy *policy = (SimPolicy *) list->data;
      g_object_unref (policy);

      list = list->next;
    }
  g_list_free (container->_priv->policies);
}

/*
 *
 *
 *
 */
SimPolicy*
sim_container_get_policy_match_ul (SimContainer     *container,
				   gint              date,
				   GInetAddr        *src_ip,
				   GInetAddr        *dst_ip,
				   SimPortProtocol  *port,
				   const gchar      *signature)
{
  SimPolicy  *policy;
  GList      *list;
  gboolean    found = FALSE;

  g_return_val_if_fail (container != NULL, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (src_ip != NULL, NULL);
  g_return_val_if_fail (dst_ip != NULL, NULL);
  g_return_val_if_fail (port != NULL, NULL);
  g_return_val_if_fail (signature != NULL, NULL);

  list = container->_priv->policies;
  while (list)
    {
      policy = (SimPolicy *) list->data;

      if (sim_policy_match (policy, date, src_ip, dst_ip, port, signature))
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return FALSE;

  return policy;
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_load_policies (SimContainer  *container,
				SimDatabase   *database)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  G_LOCK (s_mutex_policies);
  sim_container_db_load_policies_ul (container, database);
  G_UNLOCK (s_mutex_policies);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_policy (SimContainer  *container,
			   SimPolicy       *policy)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (policy);
  g_return_if_fail (SIM_IS_POLICY (policy));

  G_LOCK (s_mutex_policies);
  sim_container_append_policy_ul (container, policy);
  G_UNLOCK (s_mutex_policies);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_policy (SimContainer  *container,
			   SimPolicy       *policy)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (policy);
  g_return_if_fail (SIM_IS_POLICY (policy));

  G_LOCK (s_mutex_policies);
  sim_container_remove_policy_ul (container, policy);
  G_UNLOCK (s_mutex_policies);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_policies (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_policies);
  list = sim_container_get_policies_ul (container);
  G_UNLOCK (s_mutex_policies);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_policies (SimContainer  *container,
			    GList         *policies)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (policies);

  G_LOCK (s_mutex_policies);
  sim_container_set_policies_ul (container, policies);
  G_UNLOCK (s_mutex_policies);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_policies (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_policies);
  sim_container_free_policies_ul (container);
  G_UNLOCK (s_mutex_policies);
}

/*
 *
 *
 *
 */
SimPolicy*
sim_container_get_policy_match (SimContainer     *container,
				gint              date,
				GInetAddr        *src_ip,
				GInetAddr        *dst_ip,
				SimPortProtocol  *port,
				const gchar      *signature)
{
  SimPolicy  *policy;

  g_return_val_if_fail (container != NULL, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (src_ip != NULL, NULL);
  g_return_val_if_fail (dst_ip != NULL, NULL);
  g_return_val_if_fail (port != NULL, NULL);
  g_return_val_if_fail (signature != NULL, NULL);

  G_LOCK (s_mutex_policies);
  policy = sim_container_get_policy_match_ul (container, date, src_ip, dst_ip, port, signature);
  G_UNLOCK (s_mutex_policies);

  return policy;
}

/*
 *
 *
 *
 *
 */
void
sim_container_load_directives_from_file_ul (SimContainer  *container,
					    const gchar   *filename)
{
  SimXmlDirective *xml_directive;
  
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (filename);

  xml_directive = sim_xml_directive_new_from_file (container, filename);
  container->_priv->directives = sim_xml_directive_get_directives (xml_directive);
  g_object_unref (xml_directive);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_directive_ul (SimContainer  *container,
				   SimDirective  *directive)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (directive);
  g_return_if_fail (SIM_IS_DIRECTIVE (directive));
  
  container->_priv->directives = g_list_append (container->_priv->directives, directive);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_directive_ul (SimContainer  *container,
				   SimDirective  *directive)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (directive);
  g_return_if_fail (SIM_IS_DIRECTIVE (directive));
  
  container->_priv->directives = g_list_remove (container->_priv->directives, directive);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_directives_ul (SimContainer  *container)
{
  GList *list;
  
  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->directives);
  
  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_directives_ul (SimContainer  *container,
				 GList         *directives)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (directives);

  container->_priv->directives = g_list_concat (container->_priv->directives, directives);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_directives_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->directives;
  while (list)
    {
      SimDirective *directive = (SimDirective *) list->data;
      g_object_unref (directive);

      list = list->next;
    }
  g_list_free (container->_priv->directives);
}

/*
 *
 *
 *
 *
 */
void
sim_container_load_directives_from_file (SimContainer  *container,
					 const gchar   *filename)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (filename);

  G_LOCK (s_mutex_directives);
  sim_container_load_directives_from_file_ul (container, filename);
  G_UNLOCK (s_mutex_directives);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_directive (SimContainer  *container,
				SimDirective  *directive)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (directive);
  g_return_if_fail (SIM_IS_DIRECTIVE (directive));

  G_LOCK (s_mutex_directives);
  sim_container_append_directive_ul (container, directive);
  G_UNLOCK (s_mutex_directives);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_directive (SimContainer  *container,
				SimDirective  *directive)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (directive);
  g_return_if_fail (SIM_IS_DIRECTIVE (directive));

  G_LOCK (s_mutex_directives);
  sim_container_remove_directive_ul (container, directive);
  G_UNLOCK (s_mutex_directives);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_directives (SimContainer  *container)
{
  GList *list;
  
  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_directives);
  list = sim_container_get_directives_ul (container);
  G_UNLOCK (s_mutex_directives);
  
  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_directives (SimContainer  *container,
			      GList         *directives)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (directives);

  G_LOCK (s_mutex_directives);
  sim_container_set_directives_ul (container, directives);
  G_UNLOCK (s_mutex_directives);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_directives (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_directives);
  sim_container_free_directives_ul (container);
  G_UNLOCK (s_mutex_directives);
}

/*
 *
 *
 *
 */
void
sim_container_db_load_host_levels_ul (SimContainer  *container,
				      SimDatabase   *database)
{
  SimHostLevel  *host_level;
  GdaDataModel  *dm;
  gint           row;
  gchar         *query = "SELECT * FROM host_qualification";

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  dm = sim_database_execute_single_command (database, query);
  if (dm)
    {
      for (row = 0; row < gda_data_model_get_n_rows (dm); row++)
	{
	  host_level  = sim_host_level_new_from_dm (dm, row);

	  container->_priv->host_levels = g_list_append (container->_priv->host_levels, host_level);
	}

      g_object_unref(dm);
    }
  else
    {
      g_message ("HOST LEVELS DATA MODEL ERROR");
    }
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_insert_host_level_ul (SimContainer  *container,
				       SimDatabase   *database,
				       SimHostLevel  *host_level)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  query = sim_host_level_get_insert_clause (host_level);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_update_host_level_ul (SimContainer  *container,
				       SimDatabase   *database,
				       SimHostLevel  *host_level)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  query = sim_host_level_get_update_clause (host_level);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_delete_host_level_ul (SimContainer  *container,
				       SimDatabase   *database,
				       SimHostLevel  *host_level)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  query = sim_host_level_get_delete_clause (host_level);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_host_level_ul (SimContainer  *container,
				    SimHostLevel  *host_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  container->_priv->host_levels = g_list_append (container->_priv->host_levels, host_level);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_host_level_ul (SimContainer  *container,
				    SimHostLevel  *host_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  container->_priv->host_levels = g_list_remove (container->_priv->host_levels, host_level);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_host_levels_ul (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->host_levels);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_host_levels_ul (SimContainer  *container,
				  GList         *host_levels)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host_levels);

  container->_priv->host_levels = g_list_concat (container->_priv->host_levels, host_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_host_levels_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->host_levels;
  while (list)
    {
      SimHostLevel *host_level = (SimHostLevel *) list->data;
      g_object_unref (host_level);

      list = list->next;
    }
  g_list_free (container->_priv->host_levels);
}

/*
 *
 *
 *
 *
 */
SimHostLevel*
sim_container_get_host_level_by_ia_ul (SimContainer  *container,
				       GInetAddr     *ia)
{
  SimHostLevel  *host_level;
  GList         *list;
  gboolean       found = FALSE;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (ia, NULL);

  list = container->_priv->host_levels;
  while (list)
    {
      host_level = (SimHostLevel *) list->data;

      if (gnet_inetaddr_noport_equal (sim_host_level_get_ia (host_level), ia))
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return NULL;

  return host_level;
}

/*
 *
 *
 *
 */
void
sim_container_set_host_levels_recovery_ul (SimContainer  *container,
					   SimDatabase   *database,
					   gint           recovery)
{
  GList           *list;
  GList           *removes = NULL;
  gint             c;
  gint             a;

  g_return_if_fail (container != NULL);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (recovery > 0);

  list = container->_priv->host_levels;
  while (list)
    {
      SimHostLevel *host_level = (SimHostLevel *) list->data;

      sim_host_level_set_recovery (host_level, recovery); /* Update Memory */

      c = sim_host_level_get_c (host_level);
      a = sim_host_level_get_a (host_level);

      if (c == 0 && a == 0)
	{
	  gchar *query = sim_host_level_get_delete_clause (host_level);
	  sim_database_execute_no_query (database, query);
	  g_free (query);

	  removes = g_list_append (removes, host_level);
	}
      else
	{
	  gchar *query = sim_host_level_get_update_clause (host_level);
	  sim_database_execute_no_query (database, query);
	  g_free (query);
	}

      list = list->next;
    }

  while (removes)
    {
      SimHostLevel *host_level = (SimHostLevel *) removes->data;

      container->_priv->host_levels = g_list_remove_all (container->_priv->host_levels, host_level);
      g_object_unref (host_level);

      removes = removes->next;
    }
  g_list_free (removes);
}

/*
 *
 *
 *
 */
void
sim_container_db_load_host_levels (SimContainer  *container,
				   SimDatabase   *database)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  G_LOCK (s_mutex_host_levels);
  sim_container_db_load_host_levels_ul (container, database);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_insert_host_level (SimContainer  *container,
				    SimDatabase   *database,
				    SimHostLevel  *host_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  G_LOCK (s_mutex_host_levels);
  sim_container_db_insert_host_level_ul (container, database, host_level);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_update_host_level (SimContainer  *container,
				    SimDatabase   *database,
				    SimHostLevel  *host_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  G_LOCK (s_mutex_host_levels);
  sim_container_db_update_host_level_ul (container, database, host_level);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_delete_host_level (SimContainer  *container,
				    SimDatabase   *database,
				    SimHostLevel  *host_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  G_LOCK (s_mutex_host_levels);
  sim_container_db_delete_host_level_ul (container, database, host_level);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_host_level (SimContainer  *container,
				 SimHostLevel  *host_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  G_LOCK (s_mutex_host_levels);
  sim_container_append_host_level_ul (container, host_level);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_host_level (SimContainer  *container,
				 SimHostLevel  *host_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host_level);
  g_return_if_fail (SIM_IS_HOST_LEVEL (host_level));

  G_LOCK (s_mutex_host_levels);
  sim_container_remove_host_level_ul (container, host_level);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_host_levels (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_host_levels);
  list = sim_container_get_host_levels_ul (container);
  G_UNLOCK (s_mutex_host_levels);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_host_levels (SimContainer  *container,
			       GList         *host_levels)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (host_levels);

  G_LOCK (s_mutex_host_levels);
  sim_container_set_host_levels_ul (container, host_levels);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_host_levels (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_host_levels);
  sim_container_free_host_levels_ul (container);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 *
 */
SimHostLevel*
sim_container_get_host_level_by_ia (SimContainer  *container,
				    GInetAddr     *ia)
{
  SimHostLevel  *host_level;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (ia, NULL);

  G_LOCK (s_mutex_host_levels);
  host_level = sim_container_get_host_level_by_ia_ul (container, ia);
  G_UNLOCK (s_mutex_host_levels);

  return host_level;
}

/*
 *
 *
 *
 */
void
sim_container_set_host_levels_recovery (SimContainer  *container,
					SimDatabase   *database,
					gint           recovery)
{
  g_return_if_fail (container != NULL);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (recovery > 0);

  G_LOCK (s_mutex_host_levels);
  sim_container_set_host_levels_recovery_ul (container, database, recovery);
  G_UNLOCK (s_mutex_host_levels);
}

/*
 *
 *
 *
 */
void
sim_container_db_load_net_levels_ul (SimContainer  *container,
				     SimDatabase   *database)
{
  SimNetLevel   *net_level;
  GdaDataModel  *dm;
  gint           row;
  gchar         *query = "SELECT * FROM net_qualification";

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  dm = sim_database_execute_single_command (database, query);
  if (dm)
    {
      for (row = 0; row < gda_data_model_get_n_rows (dm); row++)
	{
	  net_level  = sim_net_level_new_from_dm (dm, row);

	  container->_priv->net_levels = g_list_append (container->_priv->net_levels, net_level);
	}

      g_object_unref(dm);
    }
  else
    {
      g_message ("NET LEVELS DATA MODEL ERROR");
    }
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_insert_net_level_ul (SimContainer  *container,
				      SimDatabase   *database,
				      SimNetLevel   *net_level)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  query = sim_net_level_get_insert_clause (net_level);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_update_net_level_ul (SimContainer  *container,
				      SimDatabase   *database,
				      SimNetLevel   *net_level)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  query = sim_net_level_get_update_clause (net_level);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_delete_net_level_ul (SimContainer  *container,
				      SimDatabase   *database,
				      SimNetLevel   *net_level)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  query = sim_net_level_get_delete_clause (net_level);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_net_level_ul (SimContainer  *container,
				   SimNetLevel   *net_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  container->_priv->net_levels = g_list_append (container->_priv->net_levels, net_level);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_net_level_ul (SimContainer  *container,
				   SimNetLevel   *net_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  container->_priv->net_levels = g_list_remove (container->_priv->net_levels, net_level);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_net_levels_ul (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->net_levels);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_net_levels_ul (SimContainer  *container,
				 GList         *net_levels)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net_levels);

  container->_priv->net_levels = g_list_concat (container->_priv->net_levels, net_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_net_levels_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->net_levels;
  while (list)
    {
      SimNetLevel *net_level = (SimNetLevel *) list->data;
      g_object_unref (net_level);

      list = list->next;
    }
  g_list_free (container->_priv->net_levels);
}

/*
 *
 *
 *
 *
 */
SimNetLevel*
sim_container_get_net_level_by_name_ul (SimContainer  *container,
					const gchar   *name)
{
  SimNetLevel  *net_level;
  GList        *list;
  gboolean      found = FALSE;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (name, NULL);

  list = container->_priv->net_levels;
  while (list)
    {
      net_level = (SimNetLevel *) list->data;

      if (!strcmp (sim_net_level_get_name (net_level), name))
	{
	  found = TRUE;
	  break;
	}

      list = list->next;
    }

  if (!found)
    return NULL;

  return net_level;
}

/*
 *
 *
 *
 */
void
sim_container_set_net_levels_recovery_ul (SimContainer  *container,
					  SimDatabase   *database,
					  gint           recovery)
{
  GList           *list;
  GList           *removes = NULL;
  gint             c;
  gint             a;

  g_return_if_fail (container != NULL);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (recovery > 0);

  list = container->_priv->net_levels;
  while (list)
    {
      SimNetLevel *net_level = (SimNetLevel *) list->data;

      sim_net_level_set_recovery (net_level, recovery); /* Update Memory */

      c = sim_net_level_get_c (net_level);
      a = sim_net_level_get_a (net_level);

      if (c == 0 && a == 0)
	{
	  gchar *query = sim_net_level_get_update_clause (net_level);
	  sim_database_execute_no_query (database, query);
	  g_free (query);

	  /* Fix this in the PostgreSQL version */
	  //container->_priv->net_levels = g_list_remove (container->_priv->net_levels, net_level); /* Delete Container List */
	  //sim_container_db_delete_net_level (container, database, net_level); /* Delete DB */
	}
      else
	{
	  gchar *query = sim_net_level_get_update_clause (net_level);
	  sim_database_execute_no_query (database, query);
	  g_free (query);
	  //sim_container_db_update_net_level (container, database, net_level); /* Update DB */
	}

      list = list->next;
    }

  while (removes)
    {
      SimNetLevel *net_level = (SimNetLevel *) removes->data;

      container->_priv->net_levels = g_list_remove_all (container->_priv->net_levels, net_level);
      g_object_unref (net_level);

      removes = removes->next;
    }
  g_list_free (removes);
}

/*
 *
 *
 *
 */
void
sim_container_db_load_net_levels (SimContainer  *container,
				  SimDatabase   *database)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));

  G_LOCK (s_mutex_net_levels);
  sim_container_db_load_net_levels_ul (container, database);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_insert_net_level (SimContainer  *container,
				   SimDatabase   *database,
				   SimNetLevel   *net_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  G_LOCK (s_mutex_net_levels);
  sim_container_db_insert_net_level_ul (container, database, net_level);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_update_net_level (SimContainer  *container,
				   SimDatabase   *database,
				   SimNetLevel   *net_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  G_LOCK (s_mutex_net_levels);
  sim_container_db_update_net_level_ul (container, database, net_level);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_delete_net_level (SimContainer  *container,
				   SimDatabase   *database,
				   SimNetLevel   *net_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  G_LOCK (s_mutex_net_levels);
  sim_container_db_delete_net_level_ul (container, database, net_level);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_net_level (SimContainer  *container,
				SimNetLevel   *net_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  G_LOCK (s_mutex_net_levels);
  sim_container_append_net_level_ul (container, net_level);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_net_level (SimContainer  *container,
				SimNetLevel   *net_level)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net_level);
  g_return_if_fail (SIM_IS_NET_LEVEL (net_level));

  G_LOCK (s_mutex_net_levels);
  sim_container_remove_net_level_ul (container, net_level);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_net_levels (SimContainer  *container)
{
  GList *list;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_net_levels);
  list = sim_container_get_net_levels_ul (container);
  G_UNLOCK (s_mutex_net_levels);

  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_net_levels (SimContainer  *container,
			      GList         *net_levels)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (net_levels);

  G_LOCK (s_mutex_net_levels);
  sim_container_set_net_levels_ul (container, net_levels);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_net_levels (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_net_levels);
  sim_container_free_net_levels_ul (container);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
SimNetLevel*
sim_container_get_net_level_by_name (SimContainer  *container,
				     const gchar   *name)
{
  SimNetLevel  *net_level;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);
  g_return_val_if_fail (name, NULL);

  G_LOCK (s_mutex_net_levels);
  net_level = sim_container_get_net_level_by_name_ul (container, name);
  G_UNLOCK (s_mutex_net_levels);

  return net_level;
}

/*
 *
 *
 *
 */
void
sim_container_set_net_levels_recovery (SimContainer  *container,
				       SimDatabase   *database,
				       gint           recovery)
{
  g_return_if_fail (container != NULL);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (recovery > 0);

  G_LOCK (s_mutex_net_levels);
  sim_container_set_net_levels_recovery_ul (container, database, recovery);
  G_UNLOCK (s_mutex_net_levels);
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_insert_backlog_ul (SimContainer  *container,
				    SimDatabase   *database,
				    SimDirective  *backlog)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));

  query = sim_directive_backlog_get_insert_clause (backlog);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_update_backlog_ul (SimContainer  *container,
				    SimDatabase   *database,
				    SimDirective   *backlog)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));

  query = sim_directive_backlog_get_update_clause (backlog);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_delete_backlog_ul (SimContainer  *container,
				    SimDatabase   *database,
				    SimDirective   *backlog)
{
  gchar *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));

  query = sim_directive_backlog_get_delete_clause (backlog);
  sim_database_execute_no_query (database, query);
  g_free (query);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_backlog_ul (SimContainer  *container,
				 SimDirective  *backlog)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));
  
  container->_priv->backlogs = g_list_append (container->_priv->backlogs, backlog);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_backlog_ul (SimContainer  *container,
				 SimDirective  *backlog)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));
  
  container->_priv->backlogs = g_list_remove (container->_priv->backlogs, backlog);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_backlogs_ul (SimContainer  *container)
{
  GList *list;
  
  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  list = g_list_copy (container->_priv->backlogs);
  
  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_backlogs_ul (SimContainer  *container,
			       GList         *backlogs)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (backlogs);

  container->_priv->backlogs = g_list_concat (container->_priv->backlogs, backlogs);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_backlogs_ul (SimContainer  *container)
{
  GList *list;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  list = container->_priv->backlogs;
  while (list)
    {
      SimDirective *backlog = (SimDirective *) list->data;
      g_object_unref (backlog);

      list = list->next;
    }
  g_list_free (container->_priv->backlogs);
}

/*
 *
 *
 *
 *
 */
void
sim_container_db_insert_backlog (SimContainer  *container,
				 SimDatabase   *database,
				 SimDirective  *backlog)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));

  G_LOCK (s_mutex_backlogs);
  sim_container_db_insert_backlog_ul (container, database, backlog);
  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_update_backlog (SimContainer  *container,
				 SimDatabase   *database,
				 SimDirective   *backlog)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));

  G_LOCK (s_mutex_backlogs);
  sim_container_db_update_backlog_ul (container, database, backlog);
  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 *
 */
void 
sim_container_db_delete_backlog (SimContainer  *container,
				 SimDatabase   *database,
				 SimDirective   *backlog)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));

  G_LOCK (s_mutex_backlogs);
  sim_container_db_delete_backlog_ul (container, database, backlog);
  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 *
 */
void
sim_container_append_backlog (SimContainer  *container,
			      SimDirective  *backlog)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));
  
  G_LOCK (s_mutex_backlogs);
  sim_container_append_backlog_ul (container, backlog);
  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 *
 */
void
sim_container_remove_backlog (SimContainer  *container,
			      SimDirective  *backlog)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (backlog);
  g_return_if_fail (SIM_IS_DIRECTIVE (backlog));
  
  G_LOCK (s_mutex_backlogs);
  sim_container_remove_backlog_ul (container, backlog);
  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 *
 */
GList*
sim_container_get_backlogs (SimContainer  *container)
{
  GList *list;
  
  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  G_LOCK (s_mutex_backlogs);
  list = sim_container_get_backlogs_ul (container);
  G_UNLOCK (s_mutex_backlogs);
  
  return list;
}

/*
 *
 *
 *
 *
 */
void
sim_container_set_backlogs (SimContainer  *container,
			    GList         *backlogs)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (backlogs);

  G_LOCK (s_mutex_backlogs);
  sim_container_set_backlogs_ul (container, backlogs);
  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_backlogs (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  G_LOCK (s_mutex_backlogs);
  sim_container_free_backlogs_ul (container);
  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 *
 */
void
sim_container_push_message (SimContainer  *container,
			    SimMessage    *message)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (message);
  g_return_if_fail (SIM_IS_MESSAGE (message));

  g_mutex_lock (container->_priv->mutex_messages);
  g_queue_push_head (container->_priv->messages, message);
  g_cond_signal (container->_priv->cond_messages);
  g_mutex_unlock (container->_priv->mutex_messages);
}

/*
 *
 *
 *
 *
 */
SimMessage*
sim_container_pop_message (SimContainer  *container)
{
  SimMessage *message;

  g_return_val_if_fail (container, NULL);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), NULL);

  g_mutex_lock (container->_priv->mutex_messages);

  while (!g_queue_peek_tail (container->_priv->messages))
    g_cond_wait (container->_priv->cond_messages, container->_priv->mutex_messages);

  message = (SimMessage *) g_queue_pop_tail (container->_priv->messages);
  if (g_queue_is_empty (container->_priv->messages))
    {
      g_cond_free (container->_priv->cond_messages);
      container->_priv->cond_messages = g_cond_new ();
    }
  g_mutex_unlock (container->_priv->mutex_messages);

  return message;
}

/*
 *
 *
 *
 *
 */
void
sim_container_free_messages (SimContainer  *container)
{
  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));

  g_mutex_lock (container->_priv->mutex_messages);
  while (!g_queue_is_empty (container->_priv->messages))
    {
      SimMessage *message = (SimMessage *) g_queue_pop_head (container->_priv->messages);
      g_object_unref (message);
    }
  g_queue_free (container->_priv->messages);
  container->_priv->messages = g_queue_new ();
  g_mutex_unlock (container->_priv->mutex_messages);
}

/*
 *
 *
 *
 *
 */
gboolean
sim_container_is_empty_messages (SimContainer  *container)
{
  gboolean empty;

  g_return_val_if_fail (container, TRUE);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), TRUE);

  g_mutex_lock (container->_priv->mutex_messages);
  empty = g_queue_is_empty (container->_priv->messages);
  g_mutex_unlock (container->_priv->mutex_messages);

  return empty;
}

/*
 *
 *
 *
 *
 */
gint
sim_container_length_messages (SimContainer  *container)
{
  gint length;

  g_return_val_if_fail (container, 0);
  g_return_val_if_fail (SIM_IS_CONTAINER (container), 0);

  g_mutex_lock (container->_priv->mutex_messages);
  length = container->_priv->messages->length;
  g_mutex_unlock (container->_priv->mutex_messages);

  return length;
}

/*
 *
 *
 *
 *
 */
void
sim_container_time_out_backlogs (SimContainer  *container,
				 SimDatabase   *database)
{
  GList         *list;
  GList         *removes = NULL;
  GTime          time_out;
  GTime          time_last;
  gint           level;
  gboolean       matched;
  GTimeVal       curr_time;
  gchar         *query;

  g_return_if_fail (container);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database);
  g_return_if_fail (SIM_IS_DATABASE (database));
  
  g_get_current_time (&curr_time);

  G_LOCK (s_mutex_backlogs);

  list = container->_priv->backlogs;
  while (list)
    {
      SimDirective *backlog = (SimDirective *) list->data;

      time_out = sim_directive_get_time_out (backlog);
      time_last = sim_directive_get_time_last (backlog);
      level = sim_directive_get_level (backlog);
      matched = sim_directive_matched (backlog);

      if ((matched) || (curr_time.tv_sec >= (time_last + time_out)))
	{
	  if (level <= 1)
	    {
	      query = sim_directive_backlog_get_delete_clause (backlog);
	      sim_database_execute_no_query (database, query);
	      g_free (query);
	    }

	  removes = g_list_append (removes, backlog);
	}

      list = list->next;
    }

  while (removes)
    {
      SimDirective *backlog = (SimDirective *) removes->data;

      container->_priv->backlogs = g_list_remove_all (container->_priv->backlogs, backlog);
      g_object_unref (backlog);

      removes = removes->next;
    }
  g_list_free (removes);

  G_UNLOCK (s_mutex_backlogs);
}

/*
 *
 *
 *
 */
void
sim_container_correlation (SimContainer *container,
			   SimDatabase  *database,
			   SimMessage   *message)
{
  GList  *list;
  static GStaticMutex mutex = G_STATIC_MUTEX_INIT;

  g_return_if_fail (container != NULL);
  g_return_if_fail (SIM_IS_CONTAINER (container));
  g_return_if_fail (database != NULL);
  g_return_if_fail (SIM_IS_DATABASE (database));
  g_return_if_fail (message != NULL);
  g_return_if_fail (SIM_IS_MESSAGE (message));


  G_LOCK (s_mutex_backlogs);

  /* Match Directives */
  list = container->_priv->directives;
  while (list)
    {
      SimDirective *directive = (SimDirective *) list->data;

      if (sim_directive_match_rule_root_by_message (directive, message))
	{
	  SimDirective *backlog;
	  SimRule      *rule_root;
	  GNode        *node;
	  GNode        *children;
	  gchar        *query;

	  backlog = sim_directive_clone (directive);
	  node = sim_directive_get_rule_root (backlog);
	  rule_root = (SimRule *) node->data;
	  sim_rule_set_message_data (rule_root, message);

	  query = sim_directive_backlog_get_insert_clause (backlog);
	  sim_database_execute_no_query (database, query);
	  g_free (query);

	  if (!G_NODE_IS_LEAF (node))
	    {
	      children = node->children;
	      while (children)
		{
		  sim_directive_set_rule_vars (backlog, children);

		  children = children->next;
		}
	    }
	  container->_priv->backlogs = g_list_append (container->_priv->backlogs, backlog);
	}
      list = list->next;
    }

  /* Match Backlogs */
  list = container->_priv->backlogs;
  while (list)
    {
      SimDirective *backlog = (SimDirective *) list->data;

      if (sim_directive_matched (backlog))
	{
	  list = list->next;
	  continue;
	}
 
      if (sim_directive_match_rule_by_message (backlog, message))
	{ 
	  gchar        *query;

	  query = sim_directive_backlog_get_update_clause (backlog);
	  sim_database_execute_no_query (database, query);
	  g_free (query);

	  if (sim_directive_matched (backlog))
	    {
	      g_message ("Backlog Directive Matched %s", sim_directive_get_name (backlog));
	    }
	}

      list = list->next;
    }

  G_UNLOCK (s_mutex_backlogs);
}
